(*
 * coqdoc - A documentation tool for Coq
 * Copyright (C) 2002 Jean-Christophe FILLIATRE
 * 
 * This software is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License version 2, as published by the Free Software Foundation.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * 
 * See the GNU General Public License version 2 for more details
 * (enclosed in the file GPL).
 *)

(*i $Id: output.ml,v 1.19 2002/02/19 14:59:22 filliatr Exp $ i*)

open Index

(*s Target language *)

type target_language = LaTeX | HTML

let target_language = ref HTML

(*s Low level output *)

let out_channel = ref stdout
let output_is_file = ref false

let set_out_file f = 
  out_channel := open_out f;
  output_is_file := true

let close () =
  if !output_is_file then close_out !out_channel

let output_char c = Pervasives.output_char !out_channel c

let output_string s = Pervasives.output_string !out_channel s

let printf s = Printf.fprintf !out_channel s

let sprintf = Printf.sprintf

let dump_file f =
  let ch = open_in f in
  try
    while true do
      Pervasives.output_char !out_channel (input_char ch)
    done
  with End_of_file -> close_in ch

(*s Options *)

let header_trailer = ref true
let quiet = ref false
let short = ref false
let index = ref true
let multi_index = ref false
let extern_defs = ref false
let page_title = ref ""
let title = ref ""

(*s Coq keywords *)

let build_table l = 
  let h = Hashtbl.create 101 in
  List.iter (fun key ->Hashtbl.add h  key ()) l;
  function s -> try Hashtbl.find h s; true with Not_found -> false

let is_keyword = 
  build_table
    [ "Add"; "AddPath"; "Axiom"; "Chapter"; "CoFixpoint";
      "CoInductive"; "Defined"; "Definition"; 
      "End"; "Export"; "Fact"; "Fix"; "Fixpoint"; "Global"; "Grammar"; "Hint";
      "Hints"; "Hypothesis"; "Immediate"; "Implicits"; "Inductive"; 
      "Infix"; "Lemma"; 
      "Load"; "Local"; "Match"; "Mutual"; "Parameter"; "Print"; "Proof"; "Qed";
      "Record"; "Recursive"; "Remark"; "Require"; "Save"; "Scheme";
      "Section"; "Show"; "Syntactic"; "Syntax"; "Tactic"; "Theorem"; 
      "Unset"; "Variable"; "Variables"; 
      (*i (* correctness *)
      "array"; "assert"; "begin"; "do"; "done"; "else"; "end"; "if";
      "in"; "invariant"; "let"; "of"; "ref"; "state"; "then"; "variant";
      "while"; i*)
      (*i (* coq terms *)
      "with"; "Case"; "Cases"; "Prop"; "Set"; "Type"; i*)
       ]

(*s Current Coq module *)

let current_module = ref ""

let set_module m = current_module := m; page_title := m

(*s Common to both LaTeX and HTML *)

let item_level = ref 0

(*s LaTeX output *)

module Latex = struct

  (*s Latex preamble *)

  let (preamble : string Queue.t) = Queue.create ()

  let push_in_preamble s = Queue.add s preamble

  let header () =
    if !header_trailer then begin
      printf "\\documentclass[12pt]{article}\n";
      printf "\\usepackage[latin1]{inputenc}\n";
      printf "\\usepackage[T1]{fontenc}\n";
      printf "\\usepackage{fullpage}\n";
      printf "\\usepackage{coqdoc}\n";
      Queue.iter (fun s -> printf "%s\n" s) preamble;
      printf "\\begin{document}\n"
    end;
    output_string 
      "%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%\n";
    output_string
      "%% This file has been automatically generated with the command\n";
    output_string "%% ";
    Array.iter (fun s -> printf "%s " s) Sys.argv;
    printf "\n";
    output_string
      "%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%\n"

  let trailer () =
    if !header_trailer then begin
      printf "\\end{document}\n"
    end

  let char c = match c with
    | '\\' -> 
	printf "\\symbol{92}"
    | '$' | '#' | '%' | '&' | '{' | '}' | '_' -> 
	output_char '\\'; output_char c
    | '^' | '~' -> 
	output_char '\\'; output_char c; printf "{}"
    | _ -> 
	output_char c

  let latex_char = output_char

  let html_char _ = ()

  let raw_ident s =
    for i = 0 to String.length s - 1 do char s.[i] done

  let start_module () = 
    if not !short then begin
      printf "\\coqdocmodule{"; 
      raw_ident !current_module; 
      printf "}\n\n"
    end

  let start_latex_math () = output_char '$'

  let stop_latex_math () = output_char '$'

  let start_verbatim () = printf "\\begin{verbatim}\n"

  let stop_verbatim () = printf "\\end{verbatim}\n"

  let indentation n = 
    if n == 0 then 
      printf "\\noindent\n"
    else
      let space = 0.5 *. (float n) in
      printf "\\coqdocindent{%2.2fem}\n" space

  let ident s _ = 
    if is_keyword s then begin
      printf "\\coqdockw{"; raw_ident s; printf "}"
    end else begin
      printf "\\coqdocid{"; raw_ident s; printf "}"
    end

  let symbol = function
    | "*"  -> printf "\\ensuremath{\\times}"
    | "->" -> printf "\\ensuremath{\\rightarrow}"
    | "<-" -> printf "\\ensuremath{\\leftarrow}"
    | "<->" -> printf "\\ensuremath{\\leftrightarrow}"
    | "=>" -> printf "\\ensuremath{\\Rightarrow}"
    | "<=" -> printf "\\ensuremath{\\le}"
    | ">=" -> printf "\\ensuremath{\\ge}"
    | "<>" -> printf "\\ensuremath{\\not=}"
    | "~" -> printf "\\ensuremath{\\lnot}"
    | "/\\" -> printf "\\ensuremath{\\land}"
    | "\\/" -> printf "\\ensuremath{\\lor}"
    | "|-" -> printf "\\ensuremath{\\vdash}"
    | s    -> raw_ident s

  let rec reach_item_level n = 
    if !item_level < n then begin
      printf "\n\\begin{itemize}\n\\item "; incr item_level;
      reach_item_level n
    end else if !item_level > n then begin
      printf "\n\\end{itemize}\n"; decr item_level;
      reach_item_level n
    end

  let item n =
    let old_level = !item_level in
    reach_item_level n;
    if n <= old_level then printf "\n\\item "

  let stop_item () = reach_item_level 0

  let start_doc () = printf "\n\n\n\\noindent\n"

  let end_doc () = stop_item (); printf "\n\n\n"

  let start_coq () = ()

  let end_coq () = ()

  let section_kind = function
    | 1 -> "\\section{"
    | 2 -> "\\subsection{"
    | 3 -> "\\subsubsection{"
    | 4 -> "\\paragraph{"
    | _ -> assert false

  let section lev s =
    stop_item ();
    output_string (section_kind lev);
    raw_ident s;
    printf "}\n\n"

  let rule () =
    printf "\\par\n\\noindent\\hrulefill\\par\n\\noindent{}"

  let paragraph () = stop_item (); printf "\n\n\\medskip\n"

  let line_break () = printf "\\coqdoceol\n"

  let start_inline_coq () = ()

  let end_inline_coq () = ()

  let make_index () = ()

end


(*s HTML output *)

module Html = struct

  let header () =
    if !header_trailer then begin
      printf "<html>\n<head>\n";
      printf "<link rel=\"stylesheet\" href=\"style.css\" type=\"text/css\">";
      printf "<title>%s</title>\n</head>\n\n" !page_title;
      printf "<body>\n\n"
    end

  let self = "http://www.lri.fr/~filliatr/coqdoc/"

  let trailer () =
    if !index && !current_module <> "Index" then 
      printf "<hr><a href=\"index.html\">Index</a>";
    if !header_trailer then begin
      printf "<hr><font size=\"-1\">This page has been generated by ";
      printf "<a href=\"%s\">coqdoc</a></font>\n" self;
      printf "</body>\n</html>"
    end

  let start_module () = 
    if not !short then
      printf "<h1>Module %s</h1>\n\n" !current_module

  let line_break () = printf "<br>\n"

  let indentation n = for i = 1 to n do printf "&nbsp;" done

  let char = function
    | '<' -> printf "&lt;"
    | '>' -> printf "&gt;"
    | '&' -> printf "&amp;"
    | c -> output_char c

  let latex_char = char

  let html_char = output_char

  let start_latex_math () = printf "<i>"
  let stop_latex_math () = printf "</i>"

  let start_verbatim () = printf "<pre>\n"
  let stop_verbatim () = printf "</pre>\n"

  let raw_ident s = for i = 0 to String.length s - 1 do char s.[i] done

  let ident s loc = 
    if is_keyword s then begin
      printf "<code class=\"keyword\">"; 
      raw_ident s; 
      printf "</code>"
    end else 
      try
	(match Index.find !current_module loc with
	   | Def _ -> 
	       printf "<a name=\"%s\"></a>" s; raw_ident s
           | Mod (m,s') when s = s' ->
	       printf "<a href=\"%s.html\">" m; raw_ident s; printf "</a>"
	   | Ref (m,s') when s = s' -> 
	       printf "<a href=\"%s.html#%s\">" m s; raw_ident s; printf "</a>"
	   | Mod _ | Ref _ ->
	       raw_ident s)
      with Not_found ->
	raw_ident s
	  
  let symbol = raw_ident

  let rec reach_item_level n = 
    if !item_level < n then begin
      printf "\n<ul>\n<li>"; incr item_level;
      reach_item_level n
    end else if !item_level > n then begin
      printf "\n</ul>\n"; decr item_level;
      reach_item_level n
    end

  let item n =
    let old_level = !item_level in
    reach_item_level n;
    if n <= old_level then printf "\n<li>"

  let stop_item () = reach_item_level 0

  let start_coq () = printf "<code>\n"

  let end_coq () = printf "</code>\n"

  let start_doc () = 
    printf "\n<table width=\"100%%\"><tr class=\"doc\"><td>\n"

  let end_doc () = stop_item (); printf "\n</td></tr></table>\n"

  let start_inline_coq () = printf "<code>"

  let end_inline_coq () = printf "</code>"

  let section lev s =
    stop_item ();
    printf "<h%d>" lev;
    raw_ident s;
    printf "</h%d>\n" lev

  let rule () = printf "<hr>\n"

  let paragraph () = stop_item (); printf "\n<p>\n"

  let entry_type = function
    | Module -> "module"
    | Definition -> "definition"
    | Inductive -> "inductive"
    | Constructor -> "constructor"
    | Lemma -> "lemma"
    | Variable -> "variable"
    | Axiom -> "axiom"
    | TacticDefinition -> "tactic definition"

  (* make a HTML index from a list of triples (name,text,link) *)
  let index_ref i c =
    let idxc = sprintf "%s_%c" i.idx_name c in
    if !multi_index then "index_" ^ idxc ^ ".html" else "index.html#" ^ idxc

  let letter_index category idx (c,l) =
    if l <> [] then begin
      let cat = if category && idx <> "global" then "(" ^ idx ^ ")" else "" in
      printf "<a name=\"%s_%c\"></a><h2>%c %s</h2>\n" idx c c cat;
      List.iter 
	(fun (id,(text,link)) -> 
	   printf "<a href=\"%s\">%s</a> %s<br>\n" link id text) l;
      printf "<p>"
    end
		
  let all_letters i = List.iter (letter_index false i.idx_name) i.idx_entries

  let separate_index navig i =
    let idx = i.idx_name in
    let one_letter ((c,l) as cl) =
      set_out_file (sprintf "index_%s_%c.html" idx c);
      header ();
      navig ();
      printf "<hr>";
      letter_index true idx cl;
      if List.length l > 30 then begin printf "<hr>"; navig () end;
      trailer ();
      close ()
    in
    List.iter one_letter i.idx_entries

  let format_global_index =
    Index.map 
      (fun s (m,t) -> 
	 if t = Module then 
	   "[module]", m ^ ".html"
	 else
	   sprintf "[%s, in <a href=\"%s.html\">%s</a>]" (entry_type t) m m ,
	   sprintf "%s.html#%s" m s)

  let format_bytype_index = function
    | Module, idx ->
	Index.map (fun id m -> "", m ^ ".html") idx
    | (t,idx) ->
	Index.map 
	  (fun s m -> 
	     let text = sprintf "[in <a href=\"%s.html\">%s</a>]" m m in
	     (text, sprintf "%s.html#%s" m s)) idx

  let navig_one_index i =
    printf "<tr>\n<td>%s Index</td>\n" (String.capitalize i.idx_name);
    List.iter 
      (fun (c,l) -> 
	 if l <> [] then
	   printf "<td><a href=\"%s\">%c</a></td>\n" (index_ref i c) c
	 else
	   printf "<td>%c</td>\n" c)
      i.idx_entries;
    let n = i.idx_size in
    printf "<td>(%d %s)</td>\n" n (if n > 1 then "entries" else "entry");
    printf "</tr>\n"

  let navig_index il = 
    printf "<table>\n";
    List.iter navig_one_index il;
    printf "</table>\n"

  let make_index () =
    if !index then begin
      let idxl = 
	let glob,bt = Index.all_entries () in
	format_global_index glob ::
	List.map format_bytype_index bt 
      in
      let navig () = navig_index idxl in 
      set_out_file "index.html";
      current_module := "Index";
      page_title := (if !title <> "" then !title else "Index");
      header ();
      if !title <> "" then printf "<h1>%s</h1>\n" !title;
      navig ();
      if !multi_index then begin
	trailer (); 
	close ();
	List.iter (separate_index navig) idxl;
      end else begin
	let one_index i =
	  if i.idx_size > 0 then begin
	    printf "<hr>\n<h1>%s Index</h1>\n" (String.capitalize i.idx_name);
	    all_letters i
	  end
	in
	List.iter one_index idxl;
	printf "<hr>";
	navig ();
	trailer ();
	close ()
      end;
    end

end


(*s Generic output *)

let select f1 f2 x = match !target_language with LaTeX -> f1 x | HTML -> f2 x

let push_in_preamble = Latex.push_in_preamble

let header = select Latex.header Html.header
let trailer = select Latex.trailer Html.trailer

let start_module = select Latex.start_module Html.start_module

let start_doc = select Latex.start_doc Html.start_doc
let end_doc = select Latex.end_doc Html.end_doc

let start_coq = select Latex.start_coq Html.start_coq
let end_coq = select Latex.end_coq Html.end_coq

let start_inline_coq = select Latex.start_inline_coq Html.start_inline_coq
let end_inline_coq = select Latex.end_inline_coq Html.end_inline_coq

let paragraph = select Latex.paragraph Html.paragraph
let line_break = select Latex.line_break Html.line_break
let indentation = select Latex.indentation Html.indentation

let section = select Latex.section Html.section
let item = select Latex.item Html.item
let stop_item = select Latex.stop_item Html.stop_item
let rule = select Latex.rule Html.rule

let char = select Latex.char Html.char
let ident = select Latex.ident Html.ident
let symbol = select Latex.symbol Html.symbol

let latex_char = select Latex.latex_char Html.latex_char
let html_char = select Latex.html_char Html.html_char

let start_latex_math = select Latex.start_latex_math Html.start_latex_math
let stop_latex_math = select Latex.stop_latex_math Html.stop_latex_math

let start_verbatim = select Latex.start_verbatim Html.start_verbatim
let stop_verbatim = select Latex.stop_verbatim Html.stop_verbatim
let verbatim_char = output_char

let make_index = select Latex.make_index Html.make_index

